__title__ = "MindsDB"
__package_name__ = "mindsdb"
__version__ = "25.9.1.0"
__description__ = "MindsDB's AI SQL Server enables developers to build AI tools that need access to real-time data to perform their tasks"
__email__ = "jorge@mindsdb.com"
__author__ = "MindsDB Inc"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "Elastic License 2.0"
__copyright__ = "Copyright(c) 2018 MindsDB, Inc"
