__title__ = "MindsDB IBM COS Handler"
__package_name__ = "mindsdb_ibm_cos_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for IBM Cloud Object Storage"
__author__ = "Ton Hoang Nguyen (Bill)"
__github__ = "'https://github.com/mindsdb/mindsdb'"
__pypi__ = "'https://github.com/mindsdb/mindsdb'"
__license__ = "MIT"
__copyright__ = "Copyright 2024 - mindsdb"
